import pytest
from typing import Any
from aquatic.calculate_qparam import (
    calculate_uniform_qparam_from_range,
    calculate_uniform_qparam_from_range_no_zero_shift,
)


@pytest.mark.parametrize("scale_orig", [1, 0.5, 0.001])
@pytest.mark.parametrize("zero_point_orig", [0, 3, 10])
def test_ternary(scale_orig: float, zero_point_orig: int):
    scale, zero_point = calculate_uniform_qparam_from_range(
        -1 * scale_orig, 1 * scale_orig, -1 + zero_point_orig, 1 + zero_point_orig
    )
    assert scale == scale_orig
    assert zero_point == zero_point_orig


def test_8bit_quant():
    scale, zero_point = calculate_uniform_qparam_from_range(-128, 127, -128, 127)
    assert scale == 1
    assert zero_point == 0


@pytest.mark.parametrize(
    "min_value, max_value, scale_orig", [(-128, 127, 1), (-128, 0, 1), (-1, 127, 1)]
)
def test_8bit_no_zero_point(min_value, max_value, scale_orig):
    scale, zero_point = calculate_uniform_qparam_from_range_no_zero_shift(
        min_value, max_value, -128, 127
    )
    assert scale == scale_orig
    assert zero_point == 0



import aquatic
import aquatic.nn as qnn
import numpy as np
import sinabs.backend.dynapcnn.dynapcnn_layer as sdl
import sinabs.layers as sl
import torch
import torch.nn as nn
from aquatic.utils import generate_dynapcnnlayer_qconfig
import sinabs.backend.dynapcnn.dynapcnn_layer as sdl
import numpy as np
import pytest


@pytest.mark.parametrize("has_bias", [True, False])
def test_dynapcnn_to_float(has_bias):
    qconfig = aquatic.utils.generate_dynapcnnlayer_qconfig(
        n_bits_weight=8, n_bits_threshold=8, has_bias=has_bias
    )
    qdynapcnn = qnn.DynapcnnLayer(
        conv=nn.Conv2d(1, 3, 3, bias=has_bias),
        spk=sl.IAF(spike_threshold=torch.tensor(1.0), min_v_mem=-1),
        in_shape=(1, 10, 10),
        discretize=False,
        qconfig=qconfig,
    )

    dynapcnn = qdynapcnn.to_float()

    # TODO this should be .all(), but cannot use with StochasticRounding
    assert (
        dynapcnn.conv_layer.weight == qdynapcnn.quantize_params()["conv_layer.weight"]
    ).any()


@pytest.mark.parametrize("has_bias", [True, False])
def test_dynapcnn_to_int(has_bias):
    qconfig = aquatic.utils.generate_dynapcnnlayer_qconfig(
        n_bits_weight=8, n_bits_threshold=8, has_bias=has_bias
    )
    qdynapcnn = qnn.DynapcnnLayer(
        conv=nn.Conv2d(1, 3, 3, bias=has_bias),
        spk=sl.IAF(spike_threshold=torch.tensor(1.0)),
        in_shape=(1, 10, 10),
        discretize=False,
        qconfig=qconfig,
    )

    int_layer = qdynapcnn.to_int()

    orig_weight_threshold_ratio = (
        qdynapcnn.mod.conv_layer.weight / qdynapcnn.mod.spk_layer.spike_threshold
    )
    int_weight_threshold_ratio = (
        int_layer.conv_layer.weight / int_layer.spk_layer.spike_threshold
    )

    torch.testing.assert_close(
        orig_weight_threshold_ratio, int_weight_threshold_ratio, atol=0.1, rtol=0.2
    )

    # TODO cumbersome way to detect that all floats are indeed integer values
    assert np.mod(int_layer.conv_layer.weight.detach().numpy(), 1).sum() == 0


@pytest.mark.parametrize("has_bias", [True, False])
def test_dynapcnn_from_float(has_bias):
    qconfig = aquatic.utils.generate_dynapcnnlayer_qconfig(
        n_bits_weight=8, n_bits_threshold=8, has_bias=has_bias
    )

    dynapcnn = sdl.DynapcnnLayer(
        conv=nn.Conv2d(1, 3, 3, bias=has_bias),
        spk=sl.IAF(spike_threshold=torch.tensor(1.0), min_v_mem=-1),
        in_shape=(1, 10, 10),
        discretize=False,
    )
    qdynapcnn = qnn.DynapcnnLayer.from_float(mod=dynapcnn, qconfig=qconfig)

    assert qdynapcnn.qconfig == qconfig
    assert (dynapcnn.conv_layer.weight == qdynapcnn.mod.conv_layer.weight).all()
    assert dynapcnn.spk_layer.spike_threshold == qdynapcnn.mod.spk_layer.spike_threshold
    assert dynapcnn.spk_layer.min_v_mem == qdynapcnn.mod.spk_layer.min_v_mem


def test_learning():
    # Defind quantization params
    qconfig = generate_dynapcnnlayer_qconfig(
        n_bits_weight=2, n_bits_threshold=4, has_bias=False
    )

    # Define a layer
    in_shape = (2, 10, 10)
    conv_layer = torch.nn.Conv2d(2, 1, 3, bias=False)
    dcnn_layer = qnn.DynapcnnLayer(
        conv=conv_layer,
        spk=sl.IAFSqueeze(batch_size=1, spike_threshold=torch.tensor(1.5)),
        in_shape=in_shape,
        discretize=False,
        qconfig=qconfig,
    )
    print(dcnn_layer.mod)

    # Generate a random input tensor
    input_data = torch.rand((1, *in_shape))

    optim = torch.optim.Adam(dcnn_layer.parameters(), lr=1e-3)
    criterion = torch.nn.MSELoss()

    # Training loop
    for i in range(10):
        optim.zero_grad()
        dcnn_layer.zero_grad()

        out = dcnn_layer(input_data)
        loss = criterion(out.mean(), torch.tensor(1.0))
        loss.backward()
        optim.step()



import aquatic
import aquatic.nn as qnn
from aquatic.qconfig import QConfig
from aquatic.fake_quantize import FakeQuantize
from aquatic.quantizer import NoZeroShiftUniformStochasticRounding
from aquatic.observer import MinMaxObserver
import torch.nn as nn
import torch
import sinabs.layers as sl


qconfig = QConfig(
    weight=FakeQuantize.with_args(
        observer=MinMaxObserver,
        quantizer=NoZeroShiftUniformStochasticRounding(
            quant_min=-128,
            quant_max=127,
        ),
    ),
    bias=FakeQuantize.with_args(
        observer=MinMaxObserver,
        quantizer=NoZeroShiftUniformStochasticRounding(
            quant_min=-128,
            quant_max=127,
        ),
    ),
)


def test_linear_qparams():
    linear = qnn.Linear(20, 40, qconfig=qconfig)
    qparams = linear.quantize_params()

    assert (qparams["weight"] != linear.mod.weight).any()
    assert (qparams["bias"] != linear.mod.bias).any()


def test_linear_to_float():
    qlinear = qnn.Linear(20, 40, qconfig=qconfig)
    linear = qlinear.to_float()

    # TODO this should be .all(), but cannot use with StochasticRounding
    assert (linear.weight == qlinear.quantize_params()["weight"]).any()


def test_linear_from_float():
    linear = nn.Linear(2, 20)
    qlinear = qnn.Linear.from_float(mod=linear, qconfig=qconfig)

    assert qlinear.qconfig == qconfig
    assert (linear.weight == qlinear.mod.weight).all()


def test_conv_qparams():
    conv = qnn.Conv2d(1, 3, 3, qconfig=qconfig)
    qparams = conv.quantize_params()

    # TODO improve these assertions
    assert (qparams["weight"] != conv.mod.weight).any()
    assert (qparams["bias"] != conv.mod.bias).any()


def test_conv_to_float():
    qconv = qnn.Conv2d(1, 3, 3, qconfig=qconfig)
    conv = qconv.to_float()

    # TODO this should be .all(), but cannot use with StochasticRounding
    assert (conv.weight == qconv.quantize_params()["weight"]).any()


def test_conv_from_float():
    conv = nn.Conv2d(1, 3, 3)
    qconv = qnn.Conv2d.from_float(mod=conv, qconfig=qconfig)

    assert qconv.qconfig == qconfig
    assert (conv.weight == qconv.mod.weight).all()


def test_linear_backwards():
    qlinear = qnn.Linear(20, 40, qconfig=qconfig)
    rand_input = torch.rand((10, 20))
    output = qlinear(rand_input)
    output.sum().backward()


def test_conv_backwards():
    qconv = qnn.Conv2d(2, 3, 3, qconfig=qconfig)
    rand_input = torch.rand((10, 2, 3, 3))
    output = qconv(rand_input)
    output.sum().backward()



from aquatic.utils import generate_dynapcnnlayer_qconfig
import pytest


@pytest.mark.parametrize("has_bias", [True, False])
def test_generate_dynapcnnlayer_qconfig(has_bias):
    qconfig = generate_dynapcnnlayer_qconfig(
        n_bits_weight=8, n_bits_threshold=8, has_bias=has_bias
    )



from audioop import add
from copy import deepcopy
import aquatic.nn as qnn
import torch.nn as nn
import torch
import aquatic
import sinabs
import sinabs.layers as sl
import sinabs.backend.dynapcnn.dynapcnn_layer as sdl


class CNN(nn.Sequential):
    def __init__(self):
        super().__init__(
            nn.Conv2d(2, 20, 5, 1, bias=False),
            nn.ReLU(),
            sl.SumPool2d(2, 2),
            nn.Conv2d(20, 32, 5, 1, bias=False),
            nn.ReLU(),
            sl.SumPool2d(2, 2),
            nn.Conv2d(32, 128, 3, 1, bias=False),
            nn.ReLU(),
            sl.SumPool2d(2, 2),
            nn.Flatten(),
            nn.Linear(128, 500, bias=False),
            nn.ReLU(),
            nn.Linear(500, 10, bias=False),
        )


def test_convert_torch_model_to_qat_model():
    model = CNN()
    qconfig = aquatic.QConfig()
    q_model = aquatic.prepare_qat(model, qconfig=qconfig)

    assert len(q_model) == len(model)
    assert isinstance(q_model[0], qnn.Conv2d)
    assert isinstance(q_model[1], nn.ReLU)
    assert isinstance(q_model[3], qnn.Conv2d)
    assert isinstance(q_model[4], nn.ReLU)
    assert isinstance(q_model[10], qnn.Linear)


def test_convert_torch_model_to_qat_model():
    model = CNN()
    snn = sinabs.from_model(model, batch_size=1, add_spiking_output=True).spiking_model

    input_shape = (1, 28, 28)
    dynapcnn_model = sinabs.backend.dynapcnn.DynapcnnNetwork(
        snn,
        input_shape=input_shape,
        dvs_input=False,
        discretize=False,
    )

    qconfig = aquatic.utils.generate_dynapcnnlayer_qconfig(
        n_bits_weight=8, n_bits_threshold=8
    )
    q_model = aquatic.prepare_qat(dynapcnn_model, qconfig=qconfig)

    assert isinstance(q_model.sequence[0], qnn.DynapcnnLayer)
    assert isinstance(q_model.sequence[1], qnn.DynapcnnLayer)
    assert isinstance(q_model.sequence[2], qnn.DynapcnnLayer)
    assert isinstance(q_model.sequence[3], qnn.DynapcnnLayer)
    assert isinstance(q_model.sequence[4], qnn.DynapcnnLayer)
    assert q_model.sequence[0].qconfig is not q_model.sequence[1].qconfig


def test_convert_qat_model_to_torch_model():
    qconfig = aquatic.QConfig()
    qmodel = nn.Sequential(
        qnn.Conv2d(2, 8, 3, qconfig=qconfig),
        nn.ReLU(),
        qnn.Conv2d(8, 16, 3, qconfig=qconfig),
        nn.ReLU(),
        nn.Flatten(),
        qnn.Linear(100, 10, qconfig=qconfig),
    )

    model = aquatic.to_float_model(qmodel)

    assert len(qmodel) == len(model)
    assert isinstance(model[0], nn.Conv2d)
    assert isinstance(model[1], nn.ReLU)
    assert isinstance(model[2], nn.Conv2d)
    assert isinstance(model[3], nn.ReLU)
    assert isinstance(model[5], nn.Linear)


def test_convert_qat_dynapcnn_model_to_float():
    qconfig = aquatic.utils.generate_dynapcnnlayer_qconfig(
        n_bits_weight=8, n_bits_threshold=8
    )
    qmodel = nn.Sequential(
        qnn.DynapcnnLayer(
            conv=nn.Conv2d(1, 1, 3),
            spk=sl.IAF(spike_threshold=torch.tensor(1.0)),
            discretize=False,
            in_shape=(1, 28, 28),
            qconfig=deepcopy(qconfig),
        ),
        qnn.DynapcnnLayer(
            conv=nn.Conv2d(1, 1, 3),
            spk=sl.IAF(spike_threshold=torch.tensor(1.0)),
            discretize=False,
            in_shape=(1, 28, 28),
            qconfig=deepcopy(qconfig),
        ),
    )

    model = aquatic.to_float_model(qmodel)

    assert isinstance(model[0], sdl.DynapcnnLayer)
    assert isinstance(model[1], sdl.DynapcnnLayer)



import torch
import sinabs.layers as sl
from aquatic.nn.lif import LIF
from aquatic import QConfig
import pytest


def test_passthrough():
    layer = LIF(tau_mem=10.0, qconfig=QConfig())

    data = torch.rand((10, 100))

    out = layer(data)
    # print(out)


@pytest.mark.parametrize(
    "tau_mem_orig, quant_max, tau_mem_quantized",
    [(20.0, 10, 10.0), (15.5, 10, 10.0), (0.0, 10, 0), (8.0, 10, 8.0)],
)
def test_tau_quantization(
    tau_mem_orig: float, quant_max: int, tau_mem_quantized: float
):
    from aquatic.quantizer import FixedQParamsUniformStochasticRounding

    qconfig = QConfig(
        tau=FixedQParamsUniformStochasticRounding.with_args(
            quant_min=-10, quant_max=quant_max, zero_point=0, scale=1.0
        )
    )
    layer = LIF(tau_mem=tau_mem_orig, qconfig=qconfig)

    quantized_params = layer.quantize_params()
    assert quantized_params["tau_mem"] == tau_mem_quantized


@pytest.mark.parametrize(
    "tau_orig, quant_max, tau_quantized",
    [(20.0, 10, 10.0), (15.5, 10, 10.0), (0.0, 10, 0.0), (8.0, 10, 8.0)],
)
def test_tau_syn_quantization(tau_orig: float, quant_max: int, tau_quantized: float):
    from aquatic.quantizer import FixedQParamsUniformStochasticRounding

    qconfig = QConfig(
        tau=FixedQParamsUniformStochasticRounding.with_args(
            quant_min=-10, quant_max=quant_max, zero_point=0, scale=1.0
        )
    )
    layer = LIF(tau_mem=tau_orig, tau_syn=tau_orig, qconfig=qconfig)

    quantized_params = layer.quantize_params()
    assert quantized_params["tau_mem"] == tau_quantized
    assert quantized_params["tau_syn"] == tau_quantized



from sinabs.backend.dynapcnn import DynapcnnNetwork
import torch.nn as nn
import torch
import sinabs
import sinabs.layers as sl
import aquatic


def test_dynapcnn_network():
    has_bias = True
    ann = nn.Sequential(
        nn.Conv2d(2, 8, 3, bias=has_bias),
        nn.ReLU(),
        nn.AvgPool2d(2),
        nn.Conv2d(8, 16, 3, bias=has_bias),
        nn.ReLU(),
        nn.AvgPool2d(2),
        nn.Conv2d(16, 2, 3, bias=has_bias),
    )

    input_shape = (2, 28, 28)
    batch_size = 1
    snn = sinabs.from_model(
        ann, input_shape=input_shape, batch_size=batch_size, add_spiking_output=True
    ).spiking_model

    dynapcnn_network = DynapcnnNetwork(
        snn=snn, input_shape=input_shape, discretize=True
    )

    num_timesteps = 100
    rand_input = torch.rand((batch_size * num_timesteps, *input_shape))
    out = dynapcnn_network(rand_input)

    qconfig = aquatic.utils.generate_dynapcnnlayer_qconfig(
        n_bits_weight=8, n_bits_threshold=8, has_bias=has_bias
    )
    qdcnn = aquatic.prepare_qat(dynapcnn_network, qconfig=qconfig)
    qdcnn = dynapcnn_network

    out = qdcnn(rand_input)

    # Training loop
    optim = torch.optim.Adam(qdcnn.parameters())
    criterion = torch.nn.MSELoss()
    for i in range(10):
        optim.zero_grad()
        qdcnn.zero_grad()

        out = qdcnn(rand_input)
        loss = criterion(out.mean(), torch.tensor(1.0))
        loss.backward()

        optim.step()



import pytest
import torch
from typing import List
from aquatic.functional.stochastic_rounding import (
    stochastic_quantize,
)


@pytest.mark.parametrize("scale", [0.01, 0.1, 1.0])
@pytest.mark.parametrize("zero_point", [0, 10, 55, 32])
@pytest.mark.parametrize(
    "data", [torch.tensor([1.0]), torch.tensor([2.0, 5.0, -1.0]), torch.tensor([-1.0])]
)
def test_scale_exactly_divisible(data: torch.Tensor, scale: float, zero_point: int):
    int_output, fq_output = stochastic_quantize(
        data, scale=scale, zero_point=zero_point
    )
    assert (int_output == zero_point + data / scale).all()
    assert (fq_output == data).all()


@pytest.mark.parametrize("scale", [1.0, 0.1, 0.01])
@pytest.mark.parametrize("zero_point", [0, 10, 20])
def test_zero_point(zero_point: int, scale: float):
    data = torch.tensor([0.01, 0.0, 4])
    # Zero point is zero
    int_output, fq_output = stochastic_quantize(
        data, scale=scale, zero_point=zero_point
    )
    assert int_output[1] == zero_point  # Zero point should be as specified
    assert (
        fq_output[1] == 0
    )  # Zero point should have no effect on the actual zero value of the fake quantized value


def test_stochasticity(scale=1.0, zero_point=0):
    data = torch.ones(1000) * 0.5
    int_output, fq_output = stochastic_quantize(
        data, scale=scale, zero_point=zero_point
    )
    assert 1 in fq_output
    assert 0 in fq_output
    assert 1 in int_output
    assert 0 in int_output
    # This might be a risky test as this assumes that the stochastically speaking, the outcome will be within the specified hardcoded interval
    assert int_output.sum() > 0.45 * 1000
    assert int_output.sum() < 0.55 * 1000



import torch
import aquatic


def test_qconfig_init():
    from aquatic.qconfig import QConfig

    config = QConfig()

    data = torch.rand((2, 4))

    # Check the identity operation
    out = config.weight()(data)
    assert (out == data).all()



import torch
from aquatic.observer import MovingAverageMinMaxObserver


def test_avg_computation():
    observer = MovingAverageMinMaxObserver(averaging_constant=0.9)

    data = torch.ones((2, 62))
    data[0] = -128
    data[1] = 127

    for _ in range(20):
        # Update data
        observer(data)

    assert observer.min_val == -128
    assert observer.max_val == 127



#!/usr/bin/env python

from distutils.core import setup

setup(
    name="aquatic",
    version="0.0.0",
    description="A quantization-aware training library for PyTorch",
    author="Sadique Sheik, Gregor Lenz",
    author_email="sadique.sheik@synsense.ai",
    packages=["aquatic"],
)



import pytest
import torch
from typing import List
from aquatic.quantizer import (
    FixedQParamsUniformStochasticRounding,
    TauStochasticRoundingToBitShift,
    AlphaStochsticRoundingToBitShift,
)


@pytest.mark.parametrize(
    "scale, expected", [(1.0, [0, 1]), (0.1, [0.5]), (0.01, [0.5])]
)
def test_fixed_params_sr(scale: float, expected: List):
    data = torch.tensor([0.5])
    # Define quantizer with default parameters
    FSR = FixedQParamsUniformStochasticRounding.with_args(scale=scale)
    my_quantizer = FSR()

    assert my_quantizer.scale == scale
    assert my_quantizer.zero_point == 0.0

    for i in range(100):
        assert my_quantizer(data) in expected


@pytest.mark.parametrize("tau", torch.arange(1, 32))
def test_tau_bitshift_sr(tau: float):
    TauSR = TauStochasticRoundingToBitShift.with_args(quant_max=3)
    tau_quantizer = TauSR()
    assert tau_quantizer.quant_max == 3

    tau_quantized = tau_quantizer(tau)
    # The maximum value allowed by 3 bit shifts is 8
    assert tau_quantized <= 8
    if tau <= 2:
        assert tau_quantized in [1, 2]
    elif tau <= 4:
        assert tau_quantized in [2, 4]
    elif tau <= 8:
        assert tau_quantized in [4, 8]
    else:
        assert tau_quantized == 8


@pytest.mark.parametrize("alpha", torch.arange(0.4, 0.9999, 0.01))
def test_alpha_bitshift_sr(alpha: float):
    AlphaSR = AlphaStochsticRoundingToBitShift.with_args(quant_max=3)
    alpha_quantizer = AlphaSR()

    alpha_quantized = alpha_quantizer(alpha)

    if alpha <= 0.5:
        assert alpha_quantized in [0, 0.5]
    elif alpha <= 0.75:
        assert alpha_quantized in [0.5, 0.75]
    else:
        alpha == 0.75



import torch
from aquatic.calculate_qparam import calculate_qparams_lif
from aquatic.nn.dynapcnn_layer import DynapcnnLayer
from aquatic.observer import SharedObserver, SharedObserverContainer
from aquatic.qconfig import QConfig
from aquatic.fake_quantize import FakeQuantize
from aquatic.quantizer import NoZeroShiftUniformStochasticRounding
import sinabs.layers as sl
import sinabs


def test_shared_observer_instantiation():
    container = SharedObserverContainer()

    weight_observer = SharedObserver.with_args(var_name="weight", container=container)()
    threshold_observer = SharedObserver.with_args(
        var_name="spike_threshold", container=container
    )()

    weight_rand = torch.rand(20)
    threshold = torch.tensor(1.0)

    weight_observer(weight_rand)
    threshold_observer(threshold)

    assert "weight" in container.states
    assert "spike_threshold" in container.states

    # print(weight_observer)


def test_qconfig_shared():
    container = SharedObserverContainer()
    weight_bitwidth = 8
    spike_threshold_bitwidth = 8
    qconfig = QConfig(
        weight=FakeQuantize.with_args(
            observer=SharedObserver.with_args(var_name="weight", container=container),
            quantizer=NoZeroShiftUniformStochasticRounding(
                quant_max=2 ** (weight_bitwidth - 1) - 1,
                quant_min=-(2 ** (weight_bitwidth - 1)),
            ),
            qparam_calculator=lambda x, y: (1, 0),
        ),
        spike_threshold=FakeQuantize.with_args(
            observer=SharedObserver.with_args(
                var_name="spike_threshold", container=container
            ),
            quantizer=NoZeroShiftUniformStochasticRounding(
                quant_max=2**spike_threshold_bitwidth - 1,
                quant_min=-(2**spike_threshold_bitwidth),
            ),
            qparam_calculator=lambda x, y: (1, 0),
        ),
    )

    dcnn_layer = DynapcnnLayer(
        conv=torch.nn.Conv2d(1, 1, 3, bias=False),
        spk=sl.IAFSqueeze(
            spike_fn=sinabs.activation.MultiSpike,
            batch_size=1,
            spike_threshold=torch.tensor(1.5),
        ),
        in_shape=(2, 10, 10),
        discretize=False,
        qconfig=qconfig,
    )

    quantized_params = dcnn_layer.quantize_params()

    assert "weight" in container.observer_handles
    assert "spike_threshold" in container.observer_handles



import sinabs
import sinabs.layers as sl
import torch
from aquatic.calculate_qparam import calculate_qparams_lif
from aquatic.fake_quantize import FakeQuantize
from aquatic.nn.dynapcnn_layer import DynapcnnLayer
from aquatic.observer import SharedObserver, SharedObserverContainer
from aquatic.qconfig import QConfig
from aquatic.quantizer import (
    NoZeroShiftUniformStochasticRounding,
    SharedQuantizer,
    SharedQuantizerContainer,
    FixedQParamsUniformStochasticRounding,
)


def test_sharing():
    container = SharedQuantizerContainer()

    weight_quantizer = SharedQuantizer(
        quantizer=NoZeroShiftUniformStochasticRounding(quant_min=-8, quant_max=7),
        container=container,
        var_name="weight",
    )
    threshold_quantizer = SharedQuantizer(
        quantizer=NoZeroShiftUniformStochasticRounding(quant_min=-64, quant_max=63),
        container=container,
        var_name="spike_threshold",
    )

    assert "weight" in container.quantizer_handles
    assert "spike_threshold" in container.quantizer_handles


def test_calculate_qparams_lif():
    quantizer_container = SharedQuantizerContainer()
    observer_container = SharedObserverContainer()

    qconfig = QConfig(
        weight=FakeQuantize.with_args(
            observer=SharedObserver.with_args(
                var_name="weight", container=observer_container
            ),
            quantizer=SharedQuantizer(
                quantizer=NoZeroShiftUniformStochasticRounding(
                    quant_min=-128, quant_max=127
                ),
                container=quantizer_container,
                var_name="weight",
            ),
            qparam_calculator=calculate_qparams_lif,
        ),
        bias=FakeQuantize.with_args(
            observer=SharedObserver.with_args(
                var_name="bias", container=observer_container
            ),
            quantizer=SharedQuantizer(
                quantizer=NoZeroShiftUniformStochasticRounding(
                    quant_min=-128, quant_max=127
                ),
                container=quantizer_container,
                var_name="bias",
            ),
            qparam_calculator=calculate_qparams_lif,
        ),
        spike_threshold=FakeQuantize.with_args(
            observer=SharedObserver.with_args(
                var_name="spike_threshold", container=observer_container
            ),
            quantizer=SharedQuantizer(
                quantizer=NoZeroShiftUniformStochasticRounding(
                    quant_min=0, quant_max=255
                ),
                container=quantizer_container,
                var_name="spike_threshold",
            ),
            qparam_calculator=calculate_qparams_lif,
        ),
    )

    conv_layer = torch.nn.Conv2d(1, 1, 3)
    # conv_layer.weight.data = torch.rand((3,3))

    dcnn_layer = DynapcnnLayer(
        conv=conv_layer,
        spk=sl.IAFSqueeze(batch_size=1, spike_threshold=torch.tensor(1.5)),
        in_shape=(2, 10, 10),
        discretize=False,
        qconfig=qconfig,
    )

    # print(dcnn_layer.mod.spk_layer.spike_threshold)

    quantized_params = dcnn_layer.quantize_params()
    # print(quantized_params)

    # print(dcnn_layer.weight_quantize.calculate_qparams())
    # print(dcnn_layer.bias_quantize.calculate_qparams())
    # print(dcnn_layer.spike_threshold_quantize.calculate_qparams())

    int_module = dcnn_layer.to_int()
    # print(int_module.conv_layer.weight)
    # print(int_module.conv_layer.bias)
    # print(int_module.spk_layer.spike_threshold)


def test_quantize_thresholds():
    qconfig = QConfig(
        spike_threshold=FixedQParamsUniformStochasticRounding.with_args(
            quant_min=0, quant_max=255
        ),
    )

    dcnn_layer = DynapcnnLayer(
        conv=torch.nn.Conv2d(1, 1, 3),
        spk=sl.IAFSqueeze(batch_size=1, spike_threshold=torch.tensor(1.5)),
        in_shape=(2, 10, 10),
        discretize=False,
        qconfig=qconfig,
    )

    quantized_params = dcnn_layer.quantize_params()

    # print(quantized_params)
